 SUBROUTINE boundary_conds( range )

!------------------------------------------------------------------------------!
! Actual revisions:
! -----------------
! The "main" part sets conditions for time level t+dt insteat of level t,
! outflow boundary conditions changed from Neumann to radiation condition
!
! Former revisions:
! -----------------
! $Id: boundary_conds.f90 73 2007-03-20 08:33:14Z raasch $
!
! 19 2007-02-23 04:53:48Z raasch
! Boundary conditions for e(nzt), pt(nzt), and q(nzt) removed because these
! gridpoints are now calculated by the prognostic equation,
! Dirichlet and zero gradient condition for pt established at top boundary
!
! RCS Log replace by Id keyword, revision history cleaned up
!
! Revision 1.15  2006/02/23 09:54:55  raasch
! Surface boundary conditions in case of topography: nzb replaced by
! 2d-k-index-arrays (nzb_w_inner, etc.). Conditions for u and v remain
! unchanged (still using nzb) because a non-flat topography must use a
! Prandtl-layer, which don't requires explicit setting of the surface values.
!
! Revision 1.1  1997/09/12 06:21:34  raasch
! Initial revision
!
!
! Description:
! ------------
! Boundary conditions for the prognostic quantities (range='main').
! In case of non-cyclic lateral boundaries the conditions for velocities at
! the outflow are set after the pressure solver has been called (range=
! 'outflow_uvw').
! One additional bottom boundary condition is applied for the TKE (=(u*)**2)
! in prandtl_fluxes. The cyclic lateral boundary conditions are implicitly
! handled in routine exchange_horiz. Pressure boundary conditions are
! explicitly set in routines pres, poisfft, poismg and sor.
!------------------------------------------------------------------------------!

    USE arrays_3d
    USE control_parameters
    USE grid_variables
    USE indices
    USE pegrid

    IMPLICIT NONE

    CHARACTER (LEN=*) ::  range

    INTEGER ::  i, j, k

    REAL    ::  c_max, c_u, c_v, c_w, denom


    IF ( range == 'main')  THEN
!
!--    Bottom boundary
       IF ( ibc_uv_b == 0 )  THEN
!
!--       Satisfying the Dirichlet condition with an extra layer below the
!--       surface where the u and v component change their sign
          u_p(nzb,:,:) = -u_p(nzb+1,:,:)
          v_p(nzb,:,:) = -v_p(nzb+1,:,:)
       ELSE
          u_p(nzb,:,:) = u_p(nzb+1,:,:)
          v_p(nzb,:,:) = v_p(nzb+1,:,:)
       ENDIF
       DO  i = nxl-1, nxr+1
          DO  j = nys-1, nyn+1
             w_p(nzb_w_inner(j,i),j,i) = 0.0
          ENDDO
       ENDDO

!
!--    Top boundary
       IF ( ibc_uv_t == 0 )  THEN
          u_p(nzt+1,:,:) = ug(nzt+1)
          v_p(nzt+1,:,:) = vg(nzt+1)
       ELSE
          u_p(nzt+1,:,:) = u_p(nzt,:,:)
          v_p(nzt+1,:,:) = v_p(nzt,:,:)
       ENDIF
       w_p(nzt:nzt+1,:,:) = 0.0  ! nzt is not a prognostic level (but cf. pres)

!
!--    Temperature at bottom boundary
       IF ( ibc_pt_b == 0 )  THEN
          DO  i = nxl-1, nxr+1
             DO  j = nys-1, nyn+1
                pt_p(nzb_s_inner(j,i),j,i) = pt(nzb_s_inner(j,i),j,i)
             ENDDO
          ENDDO
       ELSE
          DO  i = nxl-1, nxr+1
             DO  j = nys-1, nyn+1
                pt_p(nzb_s_inner(j,i),j,i) = pt_p(nzb_s_inner(j,i)+1,j,i)
             ENDDO
          ENDDO
       ENDIF

!
!--    Temperature at top boundary
       IF ( ibc_pt_t == 0 )  THEN
          pt_p(nzt+1,:,:) = pt(nzt+1,:,:)
       ELSEIF ( ibc_pt_t == 1 )  THEN
          pt_p(nzt+1,:,:) = pt_p(nzt,:,:)
       ELSEIF ( ibc_pt_t == 2 )  THEN
          pt_p(nzt+1,:,:) = pt_p(nzt,:,:)   + bc_pt_t_val * dzu(nzt+1)
       ENDIF

!
!--    Boundary conditions for TKE
!--    Generally Neumann conditions with de/dz=0 are assumed
       IF ( .NOT. constant_diffusion )  THEN
          DO  i = nxl-1, nxr+1
             DO  j = nys-1, nyn+1
                e_p(nzb_s_inner(j,i),j,i) = e_p(nzb_s_inner(j,i)+1,j,i)
             ENDDO
          ENDDO
          e_p(nzt+1,:,:) = e_p(nzt,:,:)
       ENDIF

!
!--    Boundary conditions for total water content or scalar,
!--    bottom and surface boundary (see also temperature)
       IF ( moisture  .OR.  passive_scalar )  THEN
!
!--       Surface conditions for constant_moisture_flux
          IF ( ibc_q_b == 0 ) THEN
             DO  i = nxl-1, nxr+1
                DO  j = nys-1, nyn+1
                   q_p(nzb_s_inner(j,i),j,i) = q(nzb_s_inner(j,i),j,i)
                ENDDO
             ENDDO
          ELSE
             DO  i = nxl-1, nxr+1
                DO  j = nys-1, nyn+1
                   q_p(nzb_s_inner(j,i),j,i) = q_p(nzb_s_inner(j,i)+1,j,i)
                ENDDO
             ENDDO
          ENDIF
!
!--       Top boundary
          q_p(nzt+1,:,:) = q_p(nzt,:,:)   + bc_q_t_val * dzu(nzt+1)
       ENDIF

!
!--    Lateral boundary conditions at the inflow. Quasi Neumann conditions
!--    are needed for the wall normal velocity in order to ensure zero
!--    divergence. Dirichlet conditions are used for all other quantities.
       IF ( inflow_s )  THEN
          v_p(:,nys,:) = v_p(:,nys-1,:)
       ELSEIF ( inflow_n ) THEN
          v_p(:,nyn+vynp,:) = v_p(:,nyn+vynp+1,:)
       ELSEIF ( inflow_l ) THEN
          u_p(:,:,nxl) = u_p(:,:,nxl-1)
       ELSEIF ( inflow_r ) THEN
          u_p(:,:,nxr+uxrp) = u_p(:,:,nxr+uxrp+1)
       ENDIF

!
!--    Lateral boundary conditions for scalar quantities at the outflow
       IF ( outflow_s )  THEN
          pt_p(:,nys-1,:)     = pt_p(:,nys,:)
          IF ( .NOT. constant_diffusion     )  e_p(:,nys-1,:) = e_p(:,nys,:)
          IF ( moisture .OR. passive_scalar )  q_p(:,nys-1,:) = q_p(:,nys,:)
       ELSEIF ( outflow_n )  THEN
          pt_p(:,nyn+1,:)     = pt_p(:,nyn,:)
          IF ( .NOT. constant_diffusion     )  e_p(:,nyn+1,:) = e_p(:,nyn,:)
          IF ( moisture .OR. passive_scalar )  q_p(:,nyn+1,:) = q_p(:,nyn,:)
       ELSEIF ( outflow_l )  THEN
          pt_p(:,:,nxl-1)     = pt_p(:,:,nxl)
          IF ( .NOT. constant_diffusion     )  e_p(:,:,nxl-1) = e_p(:,:,nxl)
          IF ( moisture .OR. passive_scalar )  q_p(:,:,nxl-1) = q_p(:,:,nxl)      
       ELSEIF ( outflow_r )  THEN
          pt_p(:,:,nxr+1)     = pt_p(:,:,nxr)
          IF ( .NOT. constant_diffusion     )  e_p(:,:,nxr+1) = e_p(:,:,nxr)
          IF ( moisture .OR. passive_scalar )  q_p(:,:,nxr+1) = q_p(:,:,nxr)
       ENDIF

    ENDIF

    IF ( range == 'outflow_uvw' ) THEN
!
!--    Radiation boundary condition for the velocities at the respective outflow
       IF ( outflow_s ) THEN
!          v(:,nys-1,:) = v(:,nys,:)
!          w(:,nys-1,:) = 0.0
!!
!!--       Compute the mean horizontal wind parallel to and within the outflow
!!--       wall and use this as boundary condition for u
!#if defined( __parallel )
!          CALL MPI_ALLREDUCE( uvmean_outflow_l, uvmean_outflow, nzt-nzb+2, &
!                              MPI_REAL, MPI_SUM, comm1dx, ierr )
!          uvmean_outflow = uvmean_outflow / ( nx + 1.0 )
!#else
!          uvmean_outflow = uvmean_outflow_l / ( nx + 1.0 )
!#endif
!          DO  k = nzb, nzt+1
!             u(k,nys-1,:) = uvmean_outflow(k)
!          ENDDO
       ENDIF

       IF ( outflow_n  .AND.                                                 &
            intermediate_timestep_count == intermediate_timestep_count_max ) &
       THEN

          c_max = dy / dt_3d

          DO i = nxl-1, nxr+1
             DO k = nzb+1, nzt+1

!
!--             First calculate the phase speeds for u,v, and w
                denom = u_m_n(k,ny,i,-2) - u_m_n(k,ny-1,i,-2)

                IF ( denom /= 0.0 )  THEN
                   c_u = -c_max * ( u_m_n(k,ny,i,-1)-u_m_n(k,ny,i,-2) ) / denom
                   IF ( c_u < 0.0 )  THEN
                      c_u = 0.0
                   ELSEIF ( c_u > c_max )  THEN
                      c_u = c_max
                   ENDIF
                ELSE
                   c_u = c_max
                ENDIF

                denom = v_m_n(k,ny,i,-2) - v_m_n(k,ny-1,i,-2)

                IF ( denom /= 0.0 )  THEN
                   c_v = -c_max * ( v_m_n(k,ny,i,-1)-v_m_n(k,ny,i,-2) ) / denom
                   IF ( c_v < 0.0 )  THEN
                      c_v = 0.0
                   ELSEIF ( c_v > c_max )  THEN
                      c_v = c_max
                   ENDIF
                ELSE
                   c_v = c_max
                ENDIF

                denom = w_m_n(k,ny,i,-2) - w_m_n(k,ny-1,i,-2)

                IF ( denom /= 0.0 )  THEN
                   c_w = -c_max * ( w_m_n(k,ny,i,-1)-w_m_n(k,ny,i,-2) ) / denom
                   IF ( c_w < 0.0 )  THEN
                      c_w = 0.0
                   ELSEIF ( c_w > c_max )  THEN
                      c_w = c_max
                   ENDIF
                ELSE
                   c_w = c_max
                ENDIF

!
!--             Calculate the new velocities
                u(k,ny+1,i) = u_m_n(k,ny+1,i,-1) - dt_3d * c_u * &
                              ( u_m_n(k,ny+1,i,-1) - u_m_n(k,ny,i,-1) ) * ddy

                v(k,ny+1,i) = v_m_n(k,ny+1,i,-1) - dt_3d * c_v * &
                              ( v_m_n(k,ny+1,i,-1) - v_m_n(k,ny,i,-1) ) * ddy

                w(k,ny+1,i) = w_m_n(k,ny+1,i,-1) - dt_3d * c_w * &
                              ( w_m_n(k,ny+1,i,-1) - w_m_n(k,ny,i,-1) ) * ddy

!
!--             Swap timelevels for the next timestep
                u_m_n(k,:,i,-2) = u_m_n(k,:,i,-1)
                u_m_n(k,:,i,-1) = u(k,ny-1:ny+1,i)
                v_m_n(k,:,i,-2) = v_m_n(k,:,i,-1)
                v_m_n(k,:,i,-1) = v(k,ny-1:ny+1,i)
                w_m_n(k,:,i,-2) = w_m_n(k,:,i,-1)
                w_m_n(k,:,i,-1) = w(k,ny-1:ny+1,i)

             ENDDO
          ENDDO

!
!--       Bottom boundary at the outflow
          IF ( ibc_uv_b == 0 )  THEN
             u(nzb,ny+1,:) = -u(nzb+1,ny+1,:) 
             v(nzb,ny+1,:) = -v(nzb+1,ny+1,:)  
          ELSE                    
             u(nzb,ny+1,:) =  u(nzb+1,ny+1,:)
             v(nzb,ny+1,:) =  v(nzb+1,ny+1,:)
          ENDIF
          w(nzb,ny+1,:) = 0.0

!
!--       Top boundary at the outflow
          IF ( ibc_uv_t == 0 )  THEN
             u(nzt+1,ny+1,:) = ug(nzt+1)
             v(nzt+1,ny+1,:) = vg(nzt+1)
          ELSE
             u(nzt+1,ny+1,:) = u(nzt,nyn+1,:)
             v(nzt+1,ny+1,:) = v(nzt,nyn+1,:)
          ENDIF
          w(nzt:nzt+1,ny+1,:) = 0.0

       ENDIF

       IF ( outflow_l ) THEN
!          u(:,:,nxl-1) = u(:,:,nxl)
!          w(:,:,nxl-1) = 0.0
!
!--       Compute the mean horizontal wind parallel to and within the outflow
!--       wall and use this as boundary condition for v
!#if defined( __parallel )
!          CALL MPI_ALLREDUCE( uvmean_outflow_l, uvmean_outflow, nzt-nzb+2, &
!                              MPI_REAL, MPI_SUM, comm1dy, ierr )
!          uvmean_outflow = uvmean_outflow / ( ny + 1.0 )
!#else
!          uvmean_outflow = uvmean_outflow_l / ( ny + 1.0 )
!#endif
!          DO  k = nzb, nzt+1
!             v(k,:,nxl-1) = uvmean_outflow(k)
!          ENDDO    
!
       ENDIF

       IF ( outflow_r  .AND.                                                 &
            intermediate_timestep_count == intermediate_timestep_count_max ) &
       THEN

          c_max = dx / dt_3d

          DO j = nys-1, nyn+1
             DO k = nzb+1, nzt+1

!
!--             First calculate the phase speeds for u,v, and w
                denom = u_m_r(k,j,nx,-2) - u_m_r(k,j,nx-1,-2)

                IF ( denom /= 0.0 )  THEN
                   c_u = -c_max * ( u_m_r(k,j,nx,-1)-u_m_r(k,j,nx,-2) ) / denom
                   IF ( c_u < 0.0 )  THEN
                      c_u = 0.0
                   ELSEIF ( c_u > c_max )  THEN
                      c_u = c_max
                   ENDIF
                ELSE
                   c_u = c_max
                ENDIF

                denom = v_m_r(k,j,nx,-2) - v_m_r(k,j,nx-1,-2)

                IF ( denom /= 0.0 )  THEN
                   c_v = -c_max * ( v_m_r(k,j,nx,-1)-v_m_r(k,j,nx,-2) ) / denom
                   IF ( c_v < 0.0 )  THEN
                      c_v = 0.0
                   ELSEIF ( c_v > c_max )  THEN
                      c_v = c_max
                   ENDIF
                ELSE
                   c_v = c_max
                ENDIF

                denom = w_m_r(k,j,nx,-2) - w_m_r(k,j,nx-1,-2)

                IF ( denom /= 0.0 )  THEN
                   c_w = -c_max * ( w_m_r(k,j,nx,-1)-w_m_n(k,j,nx,-2) ) / denom
                   IF ( c_w < 0.0 )  THEN
                      c_w = 0.0
                   ELSEIF ( c_w > c_max )  THEN
                      c_w = c_max
                   ENDIF
                ELSE
                   c_w = c_max
                ENDIF

!
!--             Calculate the new velocities
                u(k,j,nx+1) = u_m_r(k,j,nx+1,-1) - dt_3d * c_u * &
                              ( u_m_r(k,j,nx+1,-1) - u_m_r(k,j,nx,-1) ) * ddx

                v(k,j,nx+1) = v_m_r(k,j,nx+1,-1) - dt_3d * c_v * &
                              ( v_m_r(k,j,nx+1,-1) - v_m_r(k,j,nx,-1) ) * ddx

                w(k,j,nx+1) = w_m_r(k,j,nx+1,-1) - dt_3d * c_w * &
                              ( w_m_r(k,j,nx+1,-1) - w_m_r(k,j,nx,-1) ) * ddx

!
!--             Swap timelevels for the next timestep
                u_m_r(k,j,:,-2) = u_m_r(k,j,:,-1)
                u_m_r(k,j,:,-1) = u(k,j,nx-1:nx+1)
                v_m_r(k,j,:,-2) = v_m_r(k,j,:,-1)
                v_m_r(k,j,:,-1) = v(k,j,nx-1:nx+1)
                w_m_r(k,j,:,-2) = w_m_r(k,j,:,-1)
                w_m_r(k,j,:,-1) = w(k,j,nx-1:nx+1)

             ENDDO
          ENDDO

!
!--       Bottom boundary at the outflow
          IF ( ibc_uv_b == 0 )  THEN
             u(nzb,ny+1,:) = -u(nzb+1,ny+1,:) 
             v(nzb,ny+1,:) = -v(nzb+1,ny+1,:)  
          ELSE                    
             u(nzb,ny+1,:) =  u(nzb+1,ny+1,:)
             v(nzb,ny+1,:) =  v(nzb+1,ny+1,:)
          ENDIF
          w(nzb,ny+1,:) = 0.0

!
!--       Top boundary at the outflow
          IF ( ibc_uv_t == 0 )  THEN
             u(nzt+1,ny+1,:) = ug(nzt+1)
             v(nzt+1,ny+1,:) = vg(nzt+1)
          ELSE
             u(nzt+1,ny+1,:) = u(nzt,nyn+1,:)
             v(nzt+1,ny+1,:) = v(nzt,nyn+1,:)
          ENDIF
          w(nzt:nzt+1,ny+1,:) = 0.0

       ENDIF

    ENDIF

 
 END SUBROUTINE boundary_conds
